from collections.abc import Callable
from typing import Any

from .types import int32
from .util import IComparer_1, IEqualityComparer_1, compare, equals, physical_hash, structural_hash


def HashIdentity_FromFunctions[T](hash_1: Callable[[T], int32], eq: Callable[[T, T], bool]) -> IEqualityComparer_1[Any]:
    class ObjectExpr0(IEqualityComparer_1[Any]):
        def Equals(self, x: T, y: T, hash_1: Any = hash_1, eq: Any = eq) -> bool:
            return eq(x, y)

        def GetHashCode(self, x_1: T | None = None, hash_1: Any = hash_1, eq: Any = eq) -> int32:
            return hash_1(x_1)

    return ObjectExpr0()


def HashIdentity_Structural[T](__unit: None = None) -> IEqualityComparer_1[Any]:
    return HashIdentity_FromFunctions(structural_hash, equals)


def HashIdentity_Reference[T](__unit: None = None) -> IEqualityComparer_1[Any]:
    def _arrow1(e: T, e_1: T) -> bool:
        return e is e_1

    return HashIdentity_FromFunctions(physical_hash, _arrow1)


def ComparisonIdentity_FromFunction[T](comparer: Callable[[T, T], int32]) -> IComparer_1[T]:
    class ObjectExpr2(IComparer_1[T]):
        def Compare(self, x: T, y: T, comparer: Any = comparer) -> int32:
            return comparer(x, y)

    return ObjectExpr2()


def ComparisonIdentity_Structural[T](__unit: None = None) -> IComparer_1[Any]:
    return ComparisonIdentity_FromFunction(compare)


__all__ = [
    "HashIdentity_FromFunctions",
    "HashIdentity_Structural",
    "HashIdentity_Reference",
    "ComparisonIdentity_FromFunction",
    "ComparisonIdentity_Structural",
]
