
import { compare, IComparer, physicalHash, nullableEquals, structuralHash, IEqualityComparer, Nullable } from "./Util.js";
import { int32 } from "./Int32.js";

export function HashIdentity_FromFunctions<T>(hash: ((arg0: T) => int32), eq: ((arg0: Nullable<T>, arg1: Nullable<T>) => boolean)): IEqualityComparer<T> {
    return {
        Equals(x: Nullable<T>, y: Nullable<T>): boolean {
            return eq(x, y);
        },
        GetHashCode(x_1: T): int32 {
            return hash(x_1) | 0;
        },
    };
}

export function HashIdentity_Structural<T>(): IEqualityComparer<T> {
    return HashIdentity_FromFunctions<T>((obj: T): int32 => (structuralHash(obj) | 0), nullableEquals);
}

export function HashIdentity_Reference<T>(): IEqualityComparer<T> {
    return HashIdentity_FromFunctions<T>((obj: T): int32 => (physicalHash(obj) | 0), nullableEquals);
}

export function ComparisonIdentity_FromFunction<T>(comparer: ((arg0: Nullable<T>, arg1: Nullable<T>) => int32)): IComparer<T> {
    return {
        Compare(x: Nullable<T>, y: Nullable<T>): int32 {
            return comparer(x, y) | 0;
        },
    };
}

export function ComparisonIdentity_Structural<T>(): IComparer<T> {
    return ComparisonIdentity_FromFunction<T>((e: Nullable<T>, e_1: Nullable<T>): int32 => (compare(e, e_1) | 0));
}

