
import { IDisposable, disposeSafe, IEqualityComparer, IComparer, structuralHash, Nullable, nullableEquals } from "./Util.js";
import { int32 } from "./Int32.js";
import { HashIdentity_Structural, ComparisonIdentity_Structural } from "./FSharp.Collections.js";
import { nonNullValue, Option } from "./Option.js";
import { FSharpChoice$2_$union, FSharpChoice$2_Choice2Of2, FSharpChoice$2_Choice1Of2 } from "./Choice.js";
import { concat } from "./String.js";
import { StringBuilder__Append_Z721C83C5 } from "./System.Text.js";

export const LanguagePrimitives_GenericEqualityComparer: any = {
    "System.Collections.IEqualityComparer.Equals541DA560"(x: Nullable<any>, y: Nullable<any>): boolean {
        return nullableEquals(x, y);
    },
    "System.Collections.IEqualityComparer.GetHashCode4E60E31B"(x_1: any): int32 {
        return structuralHash(x_1) | 0;
    },
};

export const LanguagePrimitives_GenericEqualityERComparer: any = {
    "System.Collections.IEqualityComparer.Equals541DA560"(x: Nullable<any>, y: Nullable<any>): boolean {
        return nullableEquals(x, y);
    },
    "System.Collections.IEqualityComparer.GetHashCode4E60E31B"(x_1: any): int32 {
        return structuralHash(x_1) | 0;
    },
};

export function LanguagePrimitives_FastGenericComparer<T>(): IComparer<T> {
    return ComparisonIdentity_Structural<T>();
}

export function LanguagePrimitives_FastGenericComparerFromTable<T>(): IComparer<T> {
    return ComparisonIdentity_Structural<T>();
}

export function LanguagePrimitives_FastGenericEqualityComparer<T>(): IEqualityComparer<T> {
    return HashIdentity_Structural<T>();
}

export function LanguagePrimitives_FastGenericEqualityComparerFromTable<T>(): IEqualityComparer<T> {
    return HashIdentity_Structural<T>();
}

export function Operators_Failure(message: string): Error {
    return new Error(message);
}

export function Operators_FailurePattern(exn: Error): Option<string> {
    return exn.message;
}

export function Operators_NullArg<$a>(x: string): $a {
    throw new Error(x);
}

export function Operators_Using<T extends IDisposable, R>(resource: T, action: ((arg0: T) => R)): R {
    try {
        return action(resource);
    }
    finally {
        if (resource == null) {
        }
        else {
            let copyOfStruct: T = resource;
            disposeSafe(copyOfStruct);
        }
    }
}

export function Operators_Lock<$a, $b>(_lockObj: $a, action: (() => $b)): $b {
    return action();
}

export function Operators_IsNull<T>(value: T): boolean {
    if (value == null) {
        return true;
    }
    else {
        return false;
    }
}

export function Operators_IsNotNull<T>(value: T): boolean {
    if (value == null) {
        return false;
    }
    else {
        return true;
    }
}

export function Operators_IsNullV<T extends any>(value: Nullable<T>): boolean {
    return !(value != null);
}

export function Operators_NonNull<T>(value: Nullable<T>): T {
    if (value == null) {
        throw new Error();
    }
    else {
        return value;
    }
}

export function Operators_NonNullV<T extends any>(value: Nullable<T>): T {
    if (value != null) {
        return nonNullValue(value);
    }
    else {
        throw new Error();
    }
}

export function Operators_NullMatchPattern<T>(value: Nullable<T>): FSharpChoice$2_$union<void, T> {
    if (value == null) {
        return FSharpChoice$2_Choice1Of2<void, T>(undefined);
    }
    else {
        return FSharpChoice$2_Choice2Of2<void, T>(value);
    }
}

export function Operators_NullValueMatchPattern<T extends any>(value: Nullable<T>): FSharpChoice$2_$union<void, T> {
    if (value != null) {
        return FSharpChoice$2_Choice2Of2<void, T>(nonNullValue(value));
    }
    else {
        return FSharpChoice$2_Choice1Of2<void, T>(undefined);
    }
}

export function Operators_NonNullQuickPattern<T>(value: Nullable<T>): T {
    if (value == null) {
        throw new Error();
    }
    else {
        return value;
    }
}

export function Operators_NonNullQuickValuePattern<T extends any>(value: Nullable<T>): T {
    if (value != null) {
        return nonNullValue(value);
    }
    else {
        throw new Error();
    }
}

export function Operators_WithNull<T>(value: T): T {
    return value;
}

export function Operators_WithNullV<T extends any>(value: T): T {
    return value;
}

export function Operators_NullV<T extends any>(): Nullable<T> {
    return null;
}

export function Operators_NullArgCheck<T>(argumentName: string, value: Nullable<T>): T {
    if (value == null) {
        throw new Error(concat("Value cannot be null. (Parameter \'", argumentName, ..."\')"));
    }
    else {
        return value;
    }
}

export function ExtraTopLevelOperators_LazyPattern<$a>(input: any): $a {
    return input.Value;
}

export function PrintfModule_PrintFormatToStringBuilderThen<$a, $b>(continuation: (() => $a), builder: any, format: any): $b {
    return format.cont((s: string): $a => {
        StringBuilder__Append_Z721C83C5(builder, s);
        return continuation();
    });
}

export function PrintfModule_PrintFormatToStringBuilder<$a>(builder: any, format: any): $a {
    return PrintfModule_PrintFormatToStringBuilderThen<void, $a>((): void => {
    }, builder, format);
}

